import logging
import numpy as np
import pandas as pd

logger = logging.getLogger('laff')

def sequential_findflares(data) -> list:
    logger.debug("Starting sequential_findflares()")

    final_index = len(data.flux) - 2
    n = 0
    prev_start, prev_decay = 0, 0

    FLARES = []

    while n < final_index:

        dev_start = n
        dev_count = 0
        # Run deviation check.

        if data.iloc[n+1].flux > data.iloc[n].flux:
            dev_count = 1
            while data.iloc[n+dev_count+1].flux >= data.iloc[n+dev_count].flux:
                dev_count += 1

        if dev_count >= 2:
            logger.debug(f"Possible deviation from {dev_start}->{dev_start+dev_count}")

            start_point = find_start(data, dev_start, prev_decay)
            peak_point = find_peak(data, start_point)

            if check_rise(data, start_point, peak_point):
                decay_point = find_decay(data, peak_point)

                checks = [check_noise(data, start_point, peak_point, decay_point),
                          check_above(data, start_point, decay_point)]
                logger.debug(f"Checks: {checks}")

                if all(checks):
                    FLARES.append([start_point, peak_point, decay_point])
                    n = decay_point + 1
                    prev_decay = decay_point
                    continue
            else:
                # Check failed.
                logger.debug(f"Deviation NOT passed check")
        else:
            # dev_count not greater than 2, move on.
            pass

        n += 1

    return FLARES

def find_start(data: pd.DataFrame, start: int, prev_decay: int) -> int:
    """Return flare start by looking for local minima."""

    if start < 3:
        points = data.iloc[0:3]
    else:
        points = data.iloc[start-3:start+1]
    minimum = data[data.flux == min(points.flux)].index.values[0]
    minimum = (minimum + 1) if (minimum <= prev_decay) else minimum
    logger.debug(f"Flare start found at {minimum}")

    return minimum

def find_peak(data, start):
    """
    Return flare peak by looking for local maxima.

    Starting at point {start}, look for the peak of the flare. Since this could
    be one, or many points away a moving average algorithm is used. Work out
    the average of 5 point chunks and see if this is still rising. Until the
    rise stops, continue to search. Once a decay has been found, the peak is the
    datapoint with maximum value.

    :param data: The pandas dataframe containing the lightcurve data.
    :param start: Integer position of the flare start.
    :return: Integer position of the flare peak.
    """

    chunksize = 4
    prev_chunk = data['flux'].iloc[start] # Flare start position is first point.
    next_chunk = np.average(data.iloc[start+1:start+1+chunksize].flux) # Calculate first 'next chunk'.
    i = 1

    while next_chunk > prev_chunk:
        logger.debug(f"Looking at chunk i={i} : {(start+1)+(chunksize*i)}->{(start+1+4)+(chunksize*i)}")
        # Next chunk interation.
        i += 1
        prev_chunk = next_chunk
        next_chunk = np.average(data.iloc[(start+1)+(chunksize*i):(start+1+chunksize)+(chunksize*i)].flux)
    else:
        # Data has now begin to descend so look for peak up to these points.
        # Include next_chunk in case the peak lays in this list, but is just
        # brought down as an average by remaining points.
        points = data.iloc[start:(start+1+chunksize)+(chunksize*i)]
        maximum = data[data.flux == max(points.flux)].index.values[0]

        logger.debug(f"Flare peak found at {maximum}")
    return maximum

def find_decay(data: pd.DataFrame, peak: int) -> int:
    """
    Find the end of the flare as the decay smoothes into continuum.

    Longer description.

    :param data:
    :param peak:
    :returns:
    """
    decay = peak
    condition = 0
    decaypar = 3

    def calc_grad(data: pd.DataFrame, idx1: int, idx2: int) -> int:
        """Calculate gradient between first (idx1) and second (idx2) points."""
        deltaFlux = data.iloc[idx2].flux - data.iloc[idx1].flux
        deltaTime = data.iloc[idx2].time - data.iloc[idx1].time
        return deltaFlux/deltaTime

    while condition < decaypar:

        decay += 1
        logger.debug(f'for flare {peak}') # ?
        if data.idxmax('index').time in [decay + i for i in range(-1,2)]:
            logger.debug(f"Reached end of data, automatically ending flare at {decay +1}")
            return data.idxmax('index').time

        # Calculate gradients.
        NextAlong = calc_grad(data, decay, decay+1)
        PrevAlong = calc_grad(data, decay-1, decay)
        PeakToCurrent = calc_grad(data, peak, decay)
        PeakToPrev = calc_grad(data, peak, decay-1)

        cond1 = NextAlong > PeakToCurrent # Next sequence is shallower than from peak to next current.
        cond2 = NextAlong > PrevAlong # Next grad is shallower than previous grad.
        cond3 = PeakToCurrent > PeakToPrev # Peak to next point is shallower than from peak to previous point.

        # Evaluate conditions.
        if cond1 and cond2 and cond3:
            condition += 1
        elif cond1 and cond3:
            condition += 0.5
        # else:
            # condition -= 0.5 if condition >= 0.5 else 0

    logger.debug(f"Decay end found at {decay}")

    return decay


    # once end is found we will check if the flare is 'good'
    # if flare is good, accept it and continue search -> from end + 1
    # if flare is not good, disregard and continue search from deviation + 1

def check_rise(data: pd.DataFrame, start: int, peak: int) -> bool:
    """Test the rise is significant enough."""
    if data.iloc[peak].flux > data.iloc[start].flux + (2 * data.iloc[start].flux_perr):
        logger.debug("check_rise: true")
        return True
    else:
        logger.debug("check_rise: false")
        return False


def check_noise(data: pd.DataFrame, start: int, peak: int, decay: int) -> bool:
    """Check if flare is greater than x1.75 the average noise across the flare."""
    average_noise = abs(np.average(data.iloc[start:decay].flux_perr)) + abs(np.average(data.iloc[start:decay].flux_nerr))
    flux_increase = data.iloc[peak].flux - data.iloc[start].flux
    logger.debug(f"noise: {average_noise} | delta_flux: {flux_increase}")
    return True if flux_increase > 1.75 * average_noise else False

# def check_shape(data: pd.DataFrame, start: int, peak:int, decay:int) -> bool:
    # """Check the shape of the flare."""

def check_above(data: pd.DataFrame, start: int, decay: int) -> bool:
    """Check the flare is above the (estimated) continuum."""
    start = 1 if start == 0 else start # if start is first point, don't use n-1
    decay = data.idxmax('index').time - 1 if decay == data.idxmax('index').time else decay

    slope = (data['flux'].iloc[decay+1] - data['flux'].iloc[start-1])/(data['time'].iloc[decay+1] - data['time'].iloc[start-1])
    intercept = data['flux'].iloc[start-1] - slope * data['time'].iloc[start-1]

    points_above = sum(flux > (slope*time+intercept) for flux, time in zip(data['flux'].iloc[start:decay], data['time'].iloc[start:decay]))
    num_points = len(data['flux'].iloc[start:decay])

    logger.debug(f"points above/num_points => {points_above}/{num_points} = {points_above/num_points}")
    return True if points_above/num_points >= 0.7 else False



import numpy as np
import logging

logger = logging.getLogger('laff')

def possible_flares(data):
    
    # Find deviations, or possible flares.
    deviations = find_deviations(data)
    logger.debug(f'find_deviations() - found at: {deviations}')

    starts = find_minima(data, deviations) # Refine deviations by looking for local minima, or flare starts.
    logger.debug(f'{len(starts)} flare starts identified.')

    peaks = find_maxima(data, starts) # For each flare start, find the corresponding peak.
    starts, peaks = _remove_Duplicates(data, starts, peaks) # Combine any duplicate start/peaks.


    DECAYPAR = 3
    ends = _find_end(data, starts, peaks, DECAYPAR) # For each flare peak, find the corresponding flare end.

    ### TEMP
    for start, peak, end in zip(starts, peaks, ends):
        logger.critical(f"{start} / {peak} / {end}")

    # Double check starts and peaks.
    starts, peaks, ends = fixShape(data, starts, peaks, ends)

    # Combine any overlapping flares.
    # starts, peaks, ends = combine_overlaps(data, starts, peaks, ends)

    return starts, peaks, ends

def find_deviations(data):

    deviations = []
    counter = 0

    for i in data.index[:-1]:

        if data.iloc[i+1].flux > data.iloc[i].flux:
            counter += 1
        else:
            counter = 0

        if counter == 2:
            check = data.iloc[i+1].flux + (data.iloc[i+1].flux_perr * 1.5) > data.iloc[i-1].flux
            if check:
                deviations.append(i-1)
                counter = 0
            else:
                counter = 1

    return sorted(set(deviations))

def find_minima(data, deviations):
    
    minima = []

    for deviation_index in deviations:

        if deviation_index == 0:
            startpoint = 0
            endpoint   = 1
            continue

        if deviation_index < 10:
            startpoint = 0
            endpoint = deviation_index

        else:
            startpoint = deviation_index - 10
            endpoint    = deviation_index

        points = data.iloc[startpoint:endpoint]
        minimum = data[data.flux == min(points.flux)].index.values[0]
        logger.debug(f"find_minima() - original index {deviation_index}, new index {minimum}.")
        minima.append(minimum)

    return sorted(set(minima))

def find_maxima(data, starts):

    maxima = []

    for start_index in starts:
        start_index += 1
        
        # If close to end of data.
        if abs(data.idxmax('index').time - start_index) < 30:
            startpoint = start_index
            endpoint   = data.idxmax('index').time - 1

        else:
            # Keep going until next 'chunk' is higher on average.
            prev_chunk = data['flux'].iloc[start_index]
            next_chunk = np.average(data['flux'].loc[start_index:start_index+5])
            chunkcount = 1

            while next_chunk > prev_chunk:
                chunkcount += 1
                prev_chunk = next_chunk
                next_chunk = np.average(data['flux'].loc[start_index+(chunkcount*5):start_index+(chunkcount*5)+5])

            finalrange = (chunkcount + 1) * 5

            startpoint = start_index
            endpoint   = start_index + finalrange

        points = data.iloc[startpoint:endpoint]
        maximum = data[data.flux == max(points.flux)].index.values[0]
        logger.debug(f"find_maxima() - for startidx {start_index-1}, peak found at {maximum}")
        maxima.append(maximum)

    return maxima

def _remove_Duplicates(data, startlist, peaklist):
    """
    Look for flare starts with the same peak and combine.
    
    Sometimes indices A and B are found as flare starts, and both share the same
    peak C. Hence, both A and B likely should be combined as one start, the lowest
    flux is likely to be the start. Future thought: or should it just be the
    earlier index? Which is the more general case.
    """

    unique_peaks = set()
    duplicate_peaks = []
    duplicate_index = []

    indicesToRemove = []

    for idx, peak in enumerate(peaklist):
        if peak in unique_peaks:
            duplicate_peaks.append(peak)
            duplicate_index.append(idx)
        else:
            unique_peaks.add(peak)

    unique_peaks = sorted(unique_peaks)
    duplicate_peaks = sorted(duplicate_peaks)
    duplicate_index = sorted(duplicate_index)

    for data_index, peaklist_index in zip(duplicate_peaks, duplicate_index):
        pointsToCompare = [i for i, x in enumerate(peaklist) if x == data_index]
        # points is a pair of indices in peaklist
        # each peaklist has a corresponding startlist
        # so for point a and point b, find the flux in startlist at point a and b
        # compare these two
        # whichever is the lowest flux is more likely the start
        # so we keep this index and discord the other index

        comparison = np.argmin([data.iloc[startlist[x]].flux for x in pointsToCompare])

        del pointsToCompare[comparison]

        for point in pointsToCompare:
            indicesToRemove.append(point)
    
    new_startlist = [startlist[i] for i in range(len(startlist)) if i not in indicesToRemove]
    new_peaklist = [peaklist[i] for i in range(len(peaklist)) if i not in indicesToRemove]

    return new_startlist, new_peaklist

def _find_end(data, starts, peaks, DECAYPAR):
    """
    Find the end of a flare as the decay smooths into afterglow.
    
    For each peak, start counting up through data indices. At each datapoint,
    evaluate three conditions, by calculating several gradients If we reach the next
    flare start, we end the flare here immediately.
    """
    ends = []

    for start_index, peak_index in zip(starts, peaks):

        cond_count = 0
        current_index = peak_index

        while cond_count < DECAYPAR:

            # Check if we reach next peak.
            if current_index == peak_index or current_index + 1 == peak_index:
                current_index += 1
                continue
            # Check if we reach end of data.
            if any([current_index + i for i in range(2)] == data.idxmax('index').time):
                break
            # Check if we reach next start.
            if current_index + 1 in starts:
                current_index += 1
                continue

            current_index += 1

            grad_NextAlong = _calc_grad(data, current_index, current_index+1)
            grad_PrevAlong = _calc_grad(data, current_index-1, current_index)
            grad_PeakToNext = _calc_grad(data, peak_index, current_index)
            grad_PeakToPrev = _calc_grad(data, peak_index, current_index-1)

            cond1 = grad_NextAlong > grad_PeakToNext
            cond2 = grad_NextAlong > grad_PrevAlong
            cond3 = grad_PeakToNext > grad_PeakToPrev

            if cond1 and cond2 and cond3:
                cond_count += 1
            elif cond1 and cond3:
                cond_count += 0.5

            if data['flux'].iloc[current_index] > 1.5 * data['flux'].iloc[start_index]:
                if cond_count == 0:
                    cond_count = 0
                else:
                    cond_count = DECAYPAR - 0.5

        ends.append(current_index)

    return sorted(ends)

def combine_overlaps(data, starts, peaks, ends):

    if len(starts) == 0:
        return [], [], []

    combined_starts = []
    combined_peaks = []
    combined_ends = []

    sorted_pairs = sorted(zip(starts, peaks, ends), key=lambda x: x[0])
    current_start, current_peak, current_end = sorted_pairs[0]

    for start, peak, end in sorted_pairs[1:]:
        if start <= current_end:
            if data['flux'].iloc[peak] > data['flux'].iloc[current_peak]:
                current_peak = peak
            current_end = max(current_end, end)
        else:
            combined_starts.append(current_start)
            combined_peaks.append(current_peak)
            combined_ends.append(current_end)
            current_start, current_peak, current_end = start, peak, end

    combined_starts.append(current_start)
    combined_peaks.append(current_peak)
    combined_ends.append(current_end)

    return combined_starts, combined_peaks, combined_ends

def fixShape(data, starts, peaks, ends):

    if len(starts) == 0:
        return [], [], []

    adjusted_starts, adjusted_peaks = [], []

    for startidx, peakidx, endidx in zip(starts, peaks, ends):
        # Adjust start.
        look_start = data.iloc[startidx:peakidx]
        new_startidx = data[data.flux == min(look_start.flux)].index.values[0]
        if new_startidx != startidx:
            logger.debug(f"fixShape() - Shifted start {startidx} to {new_startidx}.")   
        # Adjust peak.
        look_peak = data.iloc[startidx:endidx]
        new_peakidx = data[data.flux == max(look_peak.flux)].index.values[0]
        if new_peakidx != peakidx:
            logger.debug(f"fixShape() - Shifted peak {peakidx} to {new_peakidx}.")
    
        adjusted_starts.append(new_startidx)
        adjusted_peaks.append(new_peakidx)
    
    return adjusted_starts, adjusted_peaks, ends

### FLARE CHECKS ############################################################################

def _check_FluxIncrease(data, startidx, peakidx):
    """Check the flare increase is greater than x2 the start error."""
    check = data.iloc[peakidx].flux > (data.iloc[startidx].flux + (2 * data.iloc[startidx].flux_perr))
    return check

def _check_AverageNoise(data, startidx, peakidx, endidx):
    """Check if flare is greater than x1.75 the average noise across the flare."""
    average_noise = abs(np.average(data.iloc[startidx:endidx].flux_perr)) + abs(np.average(data.iloc[startidx:endidx].flux_nerr))
    flux_increase = min(data.iloc[peakidx].flux - data.iloc[startidx].flux, data.iloc[peakidx].flux - data.iloc[endidx].flux)
    check = flux_increase > average_noise * 2
    return check

def _check_PulseShape(data, startidx, peakidx, endidx):
    try:
        rise_phase = _calc_grad(data, startidx, peakidx, indexIsRange=True)
        rise_condition = sum(x > 0 for x in rise_phase) / len(rise_phase)
    except ZeroDivisionError:
        return True

    decay_phase = _calc_grad(data, peakidx, endidx, indexIsRange=True)
    decay_condition = sum(x < 0 for x in decay_phase) / len(decay_phase)
    
    logger.debug(f'check3 conditions are {rise_condition} & {decay_condition}')
    check = rise_condition >= 0.6 and decay_condition >= 0.4
    return check

def _check_AboveContinuum(data, startidx, peakidx, endidx):
    if startidx == 0:
        startidx = 1

    slope = (data['flux'].iloc[endidx+1] - data['flux'].iloc[startidx-1])/(data['time'].iloc[endidx+1]-data['time'].iloc[startidx-1])
    intercept = data['flux'].iloc[startidx-1] - slope * data['time'].iloc[startidx-1]

    points_above = 0

    for flux, time in zip(data['flux'].iloc[startidx:endidx], data['time'].iloc[startidx:endidx]):
        if flux > (slope * time + intercept):
            points_above += 1

    logger.debug(f'check4 points above are {points_above} over {len(data["flux"].iloc[startidx:endidx])}')
    check = points_above > len(data['flux'].iloc[startidx:endidx])/2
    return check

def _calc_grad(data, index1, index2, indexIsRange=False):

    if indexIsRange == False:
        deltaFlux = data.iloc[index2].flux - data.iloc[index1].flux
        deltaTime = data.iloc[index2].time - data.iloc[index1].time
        return deltaFlux/deltaTime

    if indexIsRange == True:

        indices = range(index1, index2)
        deltaFlux = []
        deltaTime = []
        for i in indices:
            deltaFlux.append(data.iloc[i+1].flux - data.iloc[i].flux)
            deltaTime.append(data.iloc[i+1].time - data.iloc[i].time)

        return [flx / tim for flx, tim in zip(deltaFlux, deltaTime)]
    
    else:
        raise ValueError("Parameter range should be boolean.")

#### old function in main laff.py


# def old_findFlares(data):
#     """
#     Find flares within a GRB lightcurve.

#     Longer description.
    
#     [Parameters]
#         data
#             A pandas table containing the light curve data. Columns named [time,
#             time_perr, time_nerr, flux, flux_perr, flux_nerr].
            
#     [Returns]
#         flares
#             A nested list of flare start, stop, end indices.
#     """
#     logger.debug("Starting findFlares")

#     # Check data is correct input format.
#     check_data_input(data)

#     # Cutoff late data.
#     LATE_CUTOFF = True
#     data = data[data.time < 2000] if LATE_CUTOFF else data

#     from .flarefinding import possible_flares, _check_AverageNoise, _check_FluxIncrease, _check_PulseShape, _check_AboveContinuum

#     starts, peaks, ends = possible_flares(data) # Find possible flares.

#     # Perform some checks to ensure the found flares are valid.

#     all_start, all_peak, all_end = [], [], []

#     flare_start, flare_peak, flare_end = [], [], []
#     for start, peak, end in zip(starts, peaks, ends):
#         check1 = _check_AverageNoise(data, start, peak, end)
#         check2 = _check_FluxIncrease(data, start, peak)
#         check3 = _check_PulseShape(data, start, peak, end)
#         check4 = _check_AboveContinuum(data, start, peak, end)
#         logger.debug(f"Flare {round(data['time'].iloc[start],1)}-{round(data['time'].iloc[end],1)}s checks: {check1}/{check2}/{check3}/{check4}")
        
#         logger.critical(f'{start}/{peak}/{end} : {check1} / {check2} / [{check3}] / {check4}')
#         if check1 and check2 and check4:
#         # if check1 and check2 and check3 and check4:
#             flare_start.append(int(start))
#             flare_peak.append(int(peak))
#             flare_end.append(int(end))

#         all_start.append(int(start))
#         all_peak.append(int(peak))
#         all_end.append(int(end))

#     logger.info(f"Flare finder found {len(flare_start)} flare(s).")
#     return [flare_start, flare_peak, flare_end] if len(flare_start) else False



import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import logging
import warnings

# Ignore warnings.
# from pandas.core.common import SettingWithCopyWarning
# warnings.simplefilter(action='ignore', category=SettingWithCopyWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)

from .flarefinding import sequential_findflares
from .modelling import find_intial_fit, fit_continuum_mcmc, flare_fitter, broken_powerlaw, fred_flare
from .utility import check_data_input, calculate_fit_statistics, calculate_fluence

# findFlares() -- locate the indices of flares in the lightcurve
# fitContinuum(flare_indices) -- use the indices to exclude data, then fit the continuum
# fitFlares(flares, continuum) -- use indices + continuum to fit the flares

# fitGRB() -- runs all 3 function in sequence, then does some final cleanups
#          -- final statistics of the whole fit
#          -- this is what the user should be running
#          -- outputs a dictionary with all useful statistics

#################################################################################
### LOGGER
#################################################################################

logger = logging.getLogger('laff')
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)

#################################################################################
### FIND FLARES
#################################################################################



def findFlares(data):
    logger.debug("Starting sequential_findflares()")
    check_data_input(data) # First check input format is good.

    # Cutoff late data.
    LATE_CUTOFF = True
    data = data[data.time < 2000] if LATE_CUTOFF else data

    

    # Run flare finding.
    flares = sequential_findflares(data)

    logger.info(f"Found {len(flares)} flare(s).")
    return flares if len(flares) else False

#################################################################################
### CONTINUUM FITTING
#################################################################################

def fitContinuum(data, flare_indices, count_ratio):
    logger.debug(f"Starting fitContinuum")

    

    # Remove flare data.
    if flare_indices:
        for start, _, end in flare_indices:
            data = data.drop(index=range(start, end))

    # Use ODR & AIC to find best number of powerlaw breaks.
    initial_fit, initial_fit_err, initial_fit_stats = find_intial_fit(data)
    break_number = int((len(initial_fit-2)/2)-1)

    # Try an MCMC fitting run.
    try:
        final_par, final_err = fit_continuum_mcmc(data, break_number, initial_fit, initial_fit_err)
    except ValueError:
        final_par, final_err = initial_fit, initial_fit_err
        logger.debug(f"MCMC failed, defaulting to ODR.")

    # Calculate fit statistics.
    final_fit_statistics = calculate_fit_statistics(data, broken_powerlaw, final_par)
    odr_rchisq = initial_fit_stats['rchisq']
    mcmc_rchisq = final_fit_statistics['rchisq']
    logger.debug(f'ODR rchisq: {odr_rchisq}')
    logger.debug(f'MCMC rchisq: {mcmc_rchisq}')

    # Compare MCMC and ODR fits.
    if mcmc_rchisq == 0 or mcmc_rchisq < 0.1 or mcmc_rchisq == np.inf or mcmc_rchisq == -np.inf:
        logger.debug('MCMC appears to be bad, using ODR fit.')
        final_par, final_err, final_fit_statistics = initial_fit, initial_fit_err, initial_fit_stats
    elif abs(odr_rchisq-1) < abs(mcmc_rchisq-1):
        if abs(odr_rchisq-1) < 1.3 * abs(mcmc_rchisq-1):
            logger.debug("ODR better than MCMC, using ODR fit.")
            final_par, final_err, final_fit_statistics = initial_fit, initial_fit_err, initial_fit_stats
        else:
            logger.debug("ODR better than MCMC fit, but not significantly enough.")

    slopes = final_par[:break_number+1]
    slopes_err = final_err[:break_number+1]
    breaks = final_par[break_number+1:-1]
    breaks_err = final_err[break_number+1:-1]
    normal = final_par[-1]
    normal_err = final_err[-1]

    if (data.iloc[0].time < breaks[0]) and (breaks[-1] < data.iloc[-1].time):
        breakpoints = [data.iloc[0].time, *breaks, data.iloc[-1].time]
    elif (data.iloc[0].time > breaks[0]):
        breakpoints = [data.iloc[0].time, *breaks[1:], data.iloc[-1].time]
    elif (breaks[-1].time > data.iloc[-1].time):
        breakpoints = [data.iloc[0].time, *breaks[:-1], data.iloc[-1].time]

    continuum_fluence = np.sum([calculate_fluence(broken_powerlaw, final_par, breakpoints[i], breakpoints[i+1], count_ratio) for i in range(len(breakpoints)-1)])

    return {'parameters': {
                'break_num': break_number,
                'slopes': slopes, 'slopes_err': slopes_err,
                'breaks': breaks, 'breaks_err': breaks_err,
                'normal': normal, 'normal_err': normal_err},
            'fluence': continuum_fluence,
            'fit_statistics': final_fit_statistics}

#################################################################################
### FIT FLARES
#################################################################################

def fitFlares(data, flares, continuum, count_ratio):

    if not flares:
        return False

    fitted_model = broken_powerlaw(data.time, continuum['parameters'])
    data_subtracted = data.copy()
    data_subtracted['flux'] = data.flux - fitted_model

    # Fit each flare.
    flare_fits, flare_errs = flare_fitter(data, data_subtracted, flares)

    fittedFlares = []
    for indices, par, err in zip(flares, flare_fits, flare_errs):
        times = [data.iloc[x].time for x in indices]
        fluence_rise = calculate_fluence(fred_flare, par, times[0], times[1], count_ratio)
        fluence_decay = calculate_fluence(fred_flare, par, times[1], times[2], count_ratio)
        fluence_total = fluence_rise + fluence_decay

        fittedFlares.append({'times': times, 'indices': indices, 'par': par, 'par_err': err, 'fluence': [fluence_total, fluence_rise, fluence_decay]})

    return fittedFlares

#################################################################################
### FIT GRB LIGHTCURVE
#################################################################################

def fitGRB(data, count_ratio=1):
    logger.debug(f"Starting fitGRB")
    check_data_input(data)

    flare_indices = findFlares(data) # Find flare deviations.
    continuum = fitContinuum(data, flare_indices, count_ratio) # Fit continuum.
    flares = fitFlares(data, flare_indices, continuum, count_ratio) # Fit flares.

    logger.info(f"LAFF run finished.")
    return {'flares': flares, 'continuum': continuum}

#################################################################################
### PLOTTING
#################################################################################

def plotGRB(data, fitted_grb):
    logger.info(f"Starting plotGRB.")

    # Plot lightcurve.
    logger.debug("Plotting lightcurve.")
    plt.errorbar(data.time, data.flux,
                xerr=[-data.time_nerr, data.time_perr], \
                yerr=[-data.flux_nerr, data.flux_perr], \
                marker='', linestyle='None', capsize=0, zorder=1)

    # For smooth plotting of fitted functions.
    max, min = np.log10(data['time'].iloc[0]), np.log10(data['time'].iloc[-1])
    constant_range = np.logspace(min, max, num=5000)

    # Plot continuum model.
    logger.debug('Plotting continuum model.')
    fittedContinuum = broken_powerlaw(constant_range, fitted_grb['continuum']['parameters'])
    total_model = fittedContinuum
    plt.plot(constant_range, fittedContinuum, color='c')

    # Overlay marked flares.
    if fitted_grb['flares'] is not False:
        logger.debug("Plotting flare indices and models.")

        for flare in fitted_grb['flares']:
            
            # Plot flare data.
            flare_data = data.iloc[flare['indices'][0]:flare['indices'][2]]
            plt.errorbar(flare_data.time, flare_data.flux,
                        xerr=[-flare_data.time_nerr, flare_data.time_perr], \
                        yerr=[-flare_data.flux_nerr, flare_data.flux_perr], \
                        marker='', linestyle='None', capsize=0, color='r', zorder=2)

            # Plot flare models.
            flare_model = fred_flare(constant_range, flare['par'])
            total_model += flare_model
            plt.plot(constant_range, fred_flare(constant_range, flare['par']), color='tab:green', linewidth=0.6, zorder=3)

    # Plot total model.
    logger.debug('Plotting total model.')
    plt.plot(constant_range, total_model, color='tab:orange', zorder=5)
    upper_flux, lower_flux = data['flux'].max() * 10, data['flux'].min() * 0.1
    plt.ylim(lower_flux, upper_flux)

    # Plot powerlaw breaks.
    logger.debug('Plotting powerlaw breaks.')
    for x_pos in fitted_grb['continuum']['parameters']['breaks']:
        plt.axvline(x=x_pos, color='grey', linestyle='--', linewidth=0.5, zorder=0)

    logger.info("Plotting functions done, displaying...")
    plt.loglog()
    plt.show()

    return



import pandas as pd
from astropy.table import Table, vstack
import logging

logger = logging.getLogger('laff')

def lcimport(filepath, format="online_archive"):
    """
    Import a lightcurve to a format ready for LAFF.
    
    This function takes the filepath of a GRB lightcurve and will convert it into a
    format ready to use by LAFF. If this is not used, the user input is assumed to
    be structured as X.
    
    [Parameters]
        filepath (str):
            Filepath to lightcurve data.
            
    [Returns]
        data (pandas dataframe)
            Formatted pandas datatable.
    """

    if format == "online_archive":
        data = _swift_online_archive(filepath)

    elif format == "python_query":
        data = _swift_python_query(filepath)
    
    else:
        raise ValueError("Invalid format parameter.")

    logger.info('Data import successful.')
    logger.debug(f'Data filepath: {filepath}')
        
    return data


def _swift_online_archive(data_filepath):
    qdptable = []
    i = 0

    allowed_modes = ['WTSLEW', 'WT', 'WT_incbad', 'PC_incbad',
                    'batSNR5flux', 'xrtwtslewflux', 'xrtwtflux', 'xrtpcflux_incbad']
    allowed_modes = [[item] for item in allowed_modes]

    # Import tables from qdp.
    while i < 10:
        try:
            import_table = Table.read(data_filepath, format='ascii.qdp', table_id=i)
            if import_table.meta['comments'] in allowed_modes:
                qdptable.append(import_table)
            i += 1
        except FileNotFoundError:
            raise FileNotFoundError(f"No file found at '{data_filepath}'.")
        except IndexError:
            break

    # Prepare data to pandas frame.
    data = vstack(qdptable).to_pandas()
    data = data.sort_values(by=['col1'])
    data = data.reset_index(drop=True)
    data.columns = ['time', 'time_perr', 'time_nerr', 'flux', 'flux_perr', 'flux_nerr']

    return data

def _swift_python_query(data_filepath):

    qdptable = []
    i = 0

    while i < 10:
        try:
            import_table = Table.read(data_filepath, format='ascii.qdp', table_id=i)
            qdptable.append(import_table)
            logger.debug(f'Appending i = {i}')
            i += 1
        except FileNotFoundError:
            raise FileNotFoundError(f"No file found at {data_filepath}.")
        except IndexError:
            break

    data = vstack(qdptable).to_pandas()
    data = data.sort_values(by=['col1'])
    data = data.reset_index(drop=True)
    data.columns = ['time', 'time_perr', 'time_nerr', 'flux', 'flux_perr', 'flux_nerr']

    return data


import numpy as np
import pandas as pd
import scipy.integrate as integrate
import logging

logger = logging.getLogger('laff')

PAR_NAMES_FLARE = ['t_start', 'rise', 'decay', 'amplitude']
PAR_NAMES_CONTINUUM = ['break_num', 'slopes', 'slopes_err', 'breaks', 'breaks_err', 'normal', 'normal_err']
STAT_NAMES_CONTINUUM = ['chisq', 'rchisq', 'n', 'npar', 'dof', 'deltaAIC']

def calculate_fit_statistics(data, model, params):
    
    fitted_model = model(np.array(data.time), params)
    chisq = np.sum(((data.flux - fitted_model) ** 2) / (data.flux_perr ** 2))
    
    n = len(data.time)
    npar = len(params)
    dof = n - npar
    r_chisq = chisq / dof

    deltaAIC = (2 * npar) + (n * np.log(r_chisq))
        
    return {'chisq': chisq, 'rchisq': r_chisq, 'n': n, 'npar': npar, 'dof': dof, 'deltaAIC': deltaAIC}

def check_data_input(data):

    if not isinstance(data, pd.DataFrame):
        raise TypeError(f"Invalid input data type. Should be pandas dataframe.")

    # Check column names.
    expected_columns = ['time', 'time_perr', 'time_nerr', 'flux', 'flux_perr', 'flux_nerr']
    if data.shape[1] == 4:
        data.columns = ['time', 'time_perr', 'flux', 'flux_perr']
        data['time_nerr'] = data['time_perr']
        data['flux_nerr'] = data['flux_perr']
        data.columns = expected_columns
    elif data.shape[1] == 6:
        data.columns = expected_columns
    else:
        raise ValueError(f"Expected dataframe with 4 or 6 columns - got {data.shape[1]}.")

    logger.debug('Data input is good.')

    return

def calculate_fluence(model, params, start, stop, count_flux_ratio):
    """Given some model and range, calculate the fluence. Optional count/flux ratio, default 1."""

    range = np.logspace(np.log10(start), np.log10(stop), num=2500)
    fitted_model = model(range, params)
    fluence = integrate.trapezoid(fitted_model, x=range) * count_flux_ratio
    return fluence


    


import numpy as np
import logging
from ..utility import calculate_fit_statistics

logger = logging.getLogger('laff')

#################################################################################
### FRED MODEL
#################################################################################


def fred_flare(x, params):
    x = np.array(x)
    t_start = params[0]
    rise = params[1]
    decay = params[2]
    amplitude = params[3]

    cond = x < t_start


    model = amplitude * np.sqrt(np.exp(2*(rise/decay))) * np.exp(-(rise/(x-t_start))-((x-t_start)/decay))
    model[np.where(cond)] = 0

    return model

def fred_flare_wrapper(params, x):
    return fred_flare(x, params)

def all_flares(x, params):
    x = np.array(x)

    flare_params = [params[i:i+4] for i in range(0, len(params), 4)]
    
    sum_all_flares = [0.0] * len(x)

    for flare in flare_params:
        fit_flare = fred_flare(x, flare)
        sum_all_flares = [prev + current for prev, current in zip(sum_all_flares, fit_flare)]

    return sum_all_flares
    

#################################################################################
### SCIPY.ODR FITTING
#################################################################################

from scipy.odr import ODR, Model, RealData

def flare_fitter(data, residual, flares, use_odr=False):

    logger.info("Fitting flares...")

    flareFits = []
    flareErrs = []

    for start, peak, end in flares:

        data_flare = data.copy()
        data_flare['flux'] = 0
        data_flare.loc[start:end, 'flux'] = list(residual.loc[start:end, 'flux'])
        # Parameter estimates.
        t_peak = residual['time'].iloc[peak]
        t_start = residual['time'].iloc[start]
        rise = t_peak - t_start
        decay = (residual['time'].iloc[end] - t_peak)
        amplitude = abs(residual['flux'].iloc[peak] - residual['flux'].iloc[start])
        input_par = [t_start, rise, decay, amplitude]

        # Perform fit.
        logger.debug(f"For flare indices {start}/{peak}/{end}:")
        fit_par, fit_err = odr_fitter(data_flare, input_par)
        fit_par = [abs(x) for x in fit_par]
        odr_fit_par = calculate_fit_statistics(data, fred_flare, fit_par)
        odr_rchisq = odr_fit_par['rchisq']
        logger.debug(f"ODR Par: {fit_par}")
        logger.debug(f"ODR Err: {fit_err}")

        try:
            final_par, final_err = fit_flare_mcmc(data_flare, fit_par, fit_err)
            final_fit_statistics = calculate_fit_statistics(data, fred_flare, final_par)
            mcmc_rchisq = final_fit_statistics['rchisq']

            if mcmc_rchisq == 0 or mcmc_rchisq < 0.1 or mcmc_rchisq == np.inf or mcmc_rchisq == -np.inf:
                logger.debug(f'MCMC appears to be bad, using ODR fit for flare {start}-{end}.')
                final_par, final_err, final_fit_statistics = fit_par, fit_err, odr_fit_par

            elif abs(odr_rchisq-1) < abs(mcmc_rchisq-1):
                if abs(odr_rchisq) < 1.3 * abs(mcmc_rchisq-1):
                    logger.debug(f"ODR better than MCMC for flare {start}-{end}, using ODR.")
                    final_par, final_err, final_fit_statistics = fit_par, fit_err, odr_fit_par
                else:
                    logger.debug(f"ODR better than MCMC fit for flare {start}-{end}, but not significantly enough.")

        except ValueError:
            logger.debug(f'MCMC failed - using ODR fit.')
            final_par, final_err = fit_par, fit_err

        logger.debug(f"MCMC Par: {final_par}")
        logger.debug(f"MCMC Err: {final_err}")

        # Remove from residuals.
        fitted_flare = fred_flare(data.time, final_par)
        residual['flux'] -= fitted_flare

        logger.debug("Flare complete")

        flareFits.append(list(final_par))
        flareErrs.append(list(final_err))

    logger.info("Flare fitting complete for all flares.")
    return flareFits, flareErrs

def odr_fitter(data, inputpar):
    data = RealData(data.time, data.flux, data.time_perr, data.flux_perr)  
    model = Model(fred_flare_wrapper)
    odr = ODR(data, model, beta0=inputpar)

    odr.set_job(fit_type=0)
    output = odr.run()

    if output.info != 1:
        i = 1
        while output.info != 1 and i < 100:
            output = odr.restart()
            i += 1

    return output.beta, output.sd_beta

#################################################################################
### MCMC FITTING
#################################################################################

import emcee

def fit_flare_mcmc(data, init_param, init_err):

    ndim = len(init_param)
    nwalkers = 30
    nsteps = 200

    p0 = np.zeros((nwalkers, ndim))

    guess_tstart = init_param[0]
    std_tstart = init_param[0]
    p0[:, 0] = guess_tstart + std_tstart * np.random.randn(nwalkers)

    guess_rise = init_param[1]
    std_rise = init_param[1]
    p0[:, 1] = guess_rise + std_rise * np.random.randn(nwalkers)

    guess_decay = init_param[2]
    std_decay = init_param[2]
    p0[:, 2] = guess_decay + std_decay * np.random.randn(nwalkers)

    guess_amplitude = init_param[3]
    std_ampltiude = init_param[3]
    p0[:, 3] = guess_amplitude + std_ampltiude * np.random.randn(nwalkers)

    logger.debug("Running flare MCMC...")

    sampler = emcee.EnsembleSampler(nwalkers, ndim, fl_log_posterior, \
        args=(data.time, data.flux, data.time_perr, data.flux_perr))
    sampler.run_mcmc(p0, nsteps)

    burnin = 25

    samples = sampler.chain[:, burnin:, :].reshape(-1, ndim)

    fitted_par = list(map(lambda v: np.median(v), samples.T))
    fitted_err = list(map(lambda v: np.std(v), samples.T))

    logger.debug("MCMC run completed.")

    return fitted_par, fitted_err

def fl_log_likelihood(params, x, y, x_err, y_err):
    model = fred_flare(x, params)
    chisq = np.sum(( (y-model)**2) / ((y_err)**2)) 
    log_likelihood = -0.5 * np.sum(chisq + np.log(2 * np.pi * y_err**2))
    return log_likelihood

def fl_log_prior(params, TIME_END):

    t_start = params[0]
    rise = params[1]
    decay = params[2]
    amplitude = params[3]

    if not (t_start > 0) and (t_start < TIME_END):
        return -np.inf

    if rise < 0:
        return -np.inf
    if rise > TIME_END:
        return -np.inf

    if decay < 0:
        return -np.inf
    if decay > TIME_END:
        return -np.inf

    if amplitude < 0:
        return -np.inf

    return 0.0

def fl_log_posterior(params, x, y, x_err, y_err):
    lp = fl_log_prior(params, x.iloc[-1])
    ll = fl_log_likelihood(params, x, y, x_err, y_err)
    if not np.isfinite(lp):
        return -np.inf
    return lp + ll


import numpy as np
import logging
from ..utility import calculate_fit_statistics

logger = logging.getLogger('laff')

#################################################################################
### CONTINUUM MODEL
#################################################################################

def broken_powerlaw(x, params):
    
    x = np.array(x)

    if type(params) is dict:
        n = params['break_num']
        slopes = params['slopes']
        breaks = params['breaks']
        normal = params['normal']
    elif type(params) is np.ndarray or type(params) is list:
        nparam = len(params)
        n = int((nparam-2)/2)
        slopes = params[:n+1]
        breaks = params[n+1:-1]
        normal = params[-1]
    else:
        logger.critical("Input parameter is not correct type!")
        raise TypeError(f'params is not dict/list -> {type(params)}')
    
    mask = []

    for i in range(n):
        try:
            mask.append(x > breaks[i])
        except:
            logger.critical(i, 'too many?') # what's this for again?
            pass

    if n >= 0:
        model = normal * (x**(-slopes[0]))
    if n >= 1:
        model[np.where(mask[0])] = normal * (x[np.where(mask[0])]**(-slopes[1])) * (breaks[0]**(-slopes[0]+slopes[1]))
    if n >= 2:
        model[np.where(mask[1])] = normal * (x[np.where(mask[1])]**(-slopes[2])) * (breaks[0]**(-slopes[0]+slopes[1])) * (breaks[1]**(-slopes[1]+slopes[2]))
    if n >= 3:
        model[np.where(mask[2])] = normal * (x[np.where(mask[2])]**(-slopes[3])) * (breaks[0]**(-slopes[0]+slopes[1])) * (breaks[1]**(-slopes[1]+slopes[2])) * (breaks[2]**(-slopes[2]+slopes[3]))
    if n >= 4:
        model[np.where(mask[3])] = normal * (x[np.where(mask[3])]**(-slopes[4])) * (breaks[0]**(-slopes[0]+slopes[1])) * (breaks[1]**(-slopes[1]+slopes[2])) * (breaks[2]**(-slopes[2]+slopes[3])) * (breaks[3]**(-slopes[3]+slopes[4]))
    if n >= 5:
        model[np.where(mask[4])] = normal * (x[np.where(mask[4])]**(-slopes[5])) * (breaks[0]**(-slopes[0]+slopes[1])) * (breaks[1]**(-slopes[1]+slopes[2])) * (breaks[2]**(-slopes[2]+slopes[3])) * (breaks[3]**(-slopes[3]+slopes[4])) * (breaks[4]**(-slopes[4]+slopes[5]))

    return model

def broken_powerlaw_wrapper(params, x):
    """Wrapper function for broken_powerlaw - for ODR fitting."""
    return broken_powerlaw(x, params)

#################################################################################
### SCIPY.ODR FITTING
#################################################################################

from scipy.odr import ODR, Model, RealData

def find_intial_fit(data):
    data_start, data_end = data['time'].iloc[0], data['time'].iloc[-1]
    model_fits = []

    for breaknum in range(0, 6, 1):

        # Guess parameters.
        slope_guesses = [1.0] * (breaknum+1)
        break_guesses = list(np.array(np.logspace(np.log10(data_start) * 1.1, np.log10(data_end) * 0.9, num=breaknum)))
        normal_guess  = [data['flux'].iloc[0] * data['time'].iloc[0]]
        input_par = slope_guesses + break_guesses + normal_guess

        # Perform fit.
        fit_par, fit_err = odr_fitter(data, input_par)

        # Ensure breaks are sorted.
        n = int((len(fit_par)-2)/2)
        fit_par[n+1:-1] = sorted(fit_par[n+1:-1])

        # Evaluate fit.
        fit_stats = calculate_fit_statistics(data, broken_powerlaw, fit_par)
        deltaAIC = fit_stats['deltaAIC']

        model_fits.append([fit_par, deltaAIC, fit_err, fit_stats])

    # Obtain best fitted mdel.
    best_fit, best_aic, best_err, best_stats = min(model_fits, key=lambda x: x[1])

    nparam = len(best_fit)
    n = int((nparam-2)/2)
    logger.info(f"Initial continuum fit found {n} breaks.")

    logger.debug("ODR initial fit parameters.")
    logger.debug(f'Slopes: {list(best_fit[:n+1])}')
    logger.debug(f'Breaks: {list(best_fit[n+1:-1])}')
    logger.debug(f'Normal: {best_fit[-1]}')

    return best_fit, best_err, best_stats

def odr_fitter(data, inputpar):
    data = RealData(data.time, data.flux, data.time_perr, data.flux_perr)
    model = Model(broken_powerlaw_wrapper)

    odr = ODR(data, model, beta0=inputpar)

    odr.set_job(fit_type = 0)
    output = odr.run()

    if output.info != 1:
        i = 1
        while output.info != 1 and i < 100:
            output = odr.restart()
            i += 1
            
    return output.beta, output.sd_beta

#################################################################################
### MCMC FITTING
#################################################################################

import emcee

def fit_continuum_mcmc(data, breaknum, init_param, init_err):

    ndim = 2 * breaknum + 2
    nwalkers = 25
    nsteps = 500

    nparam = len(init_param)
    n = int((nparam-2)/2)

    # Parameter priors.
    p0 = np.zeros((nwalkers, ndim))

    # Slopes.
    guess_slopes = init_param[:n+1]
    std_slopes = init_err[:n+1] / 3.4
    for i in range(0, breaknum+1):
        p0[:, i] = guess_slopes[i] + std_slopes[i] * np.random.randn(nwalkers)

    # Breaks.
    guess_breaks = init_param[n+1:-1]
    std_breaks = init_err[n+1:-1] / 3.4
    for breaknum, i in enumerate(range(breaknum+1, ndim-1)):
        p0[:, i] = guess_breaks[breaknum] + std_breaks[breaknum] * np.random.randn(nwalkers)

    # Normalisation.
    guess_norm = init_param[-1]
    std_norm = init_err[-1]
    p0[:, -1] = guess_norm + std_norm * np.random.randn(nwalkers)

    logger.info("Running continuum MCMC...")

    sampler = emcee.EnsembleSampler(nwalkers, ndim, log_posterior, \
        args=(data.time, data.flux, data.time_perr, data.flux_perr))
    try:
        sampler.run_mcmc(p0, nsteps)
    except:
        sampler.run_mcmc(p0, nsteps * 3)

    burnin = 100

    samples = sampler.chain[:, burnin:, :].reshape(-1, ndim)

    fitted_par = list(map(lambda v: np.median(v), samples.T))
    fitted_err = list(map(lambda v: np.std(v), samples.T))

    logger.info("Continuum fitting complete.")

    return fitted_par, fitted_err

def log_likelihood(params, x, y, x_err, y_err):
    model = broken_powerlaw(x, params)
    # residual = y - model
    # chi_squared = np.sum((residual/ y_err) ** 2)
    # log_likelihood = -0.5 * (len(x) * np.log(2*np.pi) + np.sum(np.log(y_err ** 2)) + chi_squared)
    chisq = np.sum(( (y-model)**2) / ((y_err)**2)) 
    log_likelihood = -0.5 * np.sum(chisq + np.log(2 * np.pi * y_err**2))
    return log_likelihood

def log_prior(params, TIME_END):

    nparam = len(params)
    n = int((nparam-2)/2)

    slopes = params[:n+1]
    breaks = params[n+1:-1]
    normal = params[-1]

    if not all(-2 < value < 4 for value in slopes):
        return -np.inf

    if any(value < 0 for value in breaks):
        return -np.inf

    if any(value > TIME_END for value in breaks):
        return -np.inf

    if (normal < 0):
        return -np.inf

    return 0.0

def log_posterior(params, x, y, x_err, y_err):
    lp = log_prior(params, x.iloc[-1])
    ll = log_likelihood(params, x, y, x_err, y_err)
    if not np.isfinite(lp):
        return -np.inf
    return lp + ll